@@ -7,7 +7,7 @@ use strict;
 
 package XML::Compile::SOAP::Daemon;
 use vars '$VERSION';
-$VERSION = '2.03';
+$VERSION = '2.04';
 
 our @ISA;   # filled-in at new().
 
@@ -71,7 +71,15 @@ sub new(@)  # not called by HTTPDaemon
 
     # Upgrade daemon, wow Perl!
     @ISA = ref $daemon;
-    (bless $daemon, $class)->init(\%args);
+    my $self = (bless $daemon, $class)->init(\%args);
+
+    $self->{accept_slow_select}
+      = exists $args{accept_slow_select} ? $args{accept_slow_select} : 1; 
+
+    $self->addWsaTable(INPUT  => $args{wsa_action_input});
+    $self->addWsaTable(OUTPUT => $args{wsa_action_output});
+    $self->addSoapAction($args{soap_action_input});
+    $self;
 }
 
 sub init($)
@@ -82,7 +90,7 @@ sub init($)
         error __x"new(support_soap} removed in 2.00";
     }
 
-    my @classes = XML::Compile::Operation->registered;
+    my @classes = XML::Compile::SOAP::Operation->registered;
     @classes   # explicit load required since 2.00
         or warning "No protocol modules loaded.  Need XML::Compile::SOAP11?";
 
@@ -136,10 +144,37 @@ sub write_to_log_hook { panic "write_to_log_hook cannot be used" }
 sub outputCharset() {shift->{output_charset}}
 
 
+sub addWsaTable($@)
+{   my ($self, $dir) = (shift, shift);
+    my $h = @_==1 ? shift : { @_ };
+    my $t = $dir eq 'INPUT'  ? ($self->{wsa_input}  ||= {})
+          : $dir eq 'OUTPUT' ? ($self->{wsa_output} ||= {})
+          : error __x("addWsaTable requires 'INPUT' or 'OUTPUT', not {got}"
+              , got => $dir);
+
+    while(my($op, $action) = each %$h) { $t->{$op} ||= $action }
+    $t;
+}
+
+
+sub addSoapAction(@)
+{   my $self = shift;
+    my $h = @_==1 ? shift : { @_ };
+    my $t = $self->{sa_input}     ||= {};
+    my $r = $self->{sa_input_rev} ||= {};
+    while(my($op, $action) = each %$h)
+    {   $t->{$op}     ||= $action;
+        $r->{$action} ||= $op;
+    }
+    $t;
+}
+
+
 sub run(@)
 {   my ($self, %args) = @_;
     delete $args{log_file};      # Net::Server should not mess with my preps
     $args{no_client_stdout} = 1; # it's a daemon, you know
+    $self->{wsa_input_rev}  = +{ reverse %{$self->{wsa_input}} };
     $self->SUPER::run(%args);
 }
 
@@ -148,7 +183,7 @@ sub run(@)
 sub process_request(@) { panic "must be extended" }
 
 sub process($)
-{   my ($self, $input) = @_;
+{   my ($self, $input, $req, $soapaction) = @_;
 
     my $xmlin;
     if(ref $input eq 'SCALAR')
@@ -157,34 +192,66 @@ sub process($)
             or return $self->faultInvalidXML($@->died)
     }
     else
-    {   $xmlin   = $input;
+    {   $xmlin = $input;
     }
     
-    $xmlin       = $xmlin->documentElement
+    $xmlin     = $xmlin->documentElement
         if $xmlin->isa('XML::LibXML::Document');
 
-    my $local    = $xmlin->localName;
+    my $local  = $xmlin->localName;
     $local eq 'Envelope'
         or return $self->faultNotSoapMessage(type_of_node $xmlin);
 
-    my $envns    = $xmlin->namespaceURI || '';
-    my $proto    = XML::Compile::Operation->fromEnvelope($envns)
+    my $envns  = $xmlin->namespaceURI || '';
+    my $proto  = XML::Compile::SOAP::Operation->fromEnvelope($envns)
         or return $self->faultUnsupportedSoapVersion($envns);
     # proto is a XML::Compile::SOAP*::Operation
-    my $server   = $proto->serverClass;
+    my $server = $proto->serverClass;
 
-    my $info     = XML::Compile::SOAP->messageStructure($xmlin);
+    my $info   = XML::Compile::SOAP->messageStructure($xmlin);
     my $version  = $info->{soap_version} = $proto->version;
     my $handlers = $self->{handler}{$version} || {};
 
-    keys %$handlers;  # reset each()
-    while(my ($name, $handler) = each %$handlers)
-    {
-        my ($rc, $msg, $xmlout) = $handler->($name, $xmlin, $info);
-        defined $xmlout or next;
+    # Try to resolve operation via WSA
+    my $wsa_in   = $self->{wsa_input_rev};
+    if(my $wsa_action = $info->{wsa_action})
+    {   if(my $name = $wsa_in->{$wsa_action})
+        {   my $handler = $handlers->{$name};
+            local $info->{selected_by} = 'wsa-action';
+            my ($rc, $msg, $xmlout) = $handler->($name, $xmlin, $info);
+            if($xmlout)
+            {   trace "data ready for $version $name, via wsa $wsa_action";
+                return ($rc, $msg, $xmlout);
+            }
+        }
+    }
 
-        trace "data ready for $version $name";
-        return ($rc, $msg, $xmlout);
+    # Try to resolve operation via soapAction
+    my $sa = $self->{sa_input_rev};
+    if(defined $soapaction && $soapaction =~ m/^\s*(["'])?(.+)\1\s*$/)
+    {   if(my $name = $sa->{$1})
+        {   my $handler = $handlers->{$name};
+            local $info->{selected_by} = 'soap-action';
+            my ($rc, $msg, $xmlout) = $handler->($name, $xmlin, $info);
+            if($xmlout)
+            {   trace "data ready for $version $name, via sa $soapaction";
+                return ($rc, $msg, $xmlout);
+            }
+        }
+    }
+
+    # Last resort, try each of the operations for the first which
+    # can be parsed correctly.
+    if($self->{accept_slow_select})
+    {   keys %$handlers;  # reset each()
+        $info->{selected_by} = 'attempt all';
+        while(my ($name, $handler) = each %$handlers)
+        {   my ($rc, $msg, $xmlout) = $handler->($name, $xmlin, $info);
+            defined $xmlout or next;
+
+            trace "data ready for $version $name";
+            return ($rc, $msg, $xmlout);
+        }
     }
 
     my $bodyel = $info->{body}[0] || '(none)';
@@ -197,16 +264,18 @@ sub process($)
 
     my @available = sort keys %$handlers;
     ( RC_NOT_FOUND, 'message not recognized'
-    , $server->faultMessageNotRecognized($bodyel, \@available));
+    , $server->faultMessageNotRecognized($bodyel, $soapaction, \@available));
 }
 
 
 sub operationsFromWSDL($@)
 {   my ($self, $wsdl, %args) = @_;
-    my %callbacks = $args{callbacks} ? %{$args{callbacks}} : ();
+    my %callbacks  = $args{callbacks} ? %{$args{callbacks}} : ();
     my %names;
 
-    my $default = $args{default_callback};
+    my $default_cb = $args{default_callback};
+    my $wsa_input  = $self->{wsa_input};
+    my $wsa_output = $self->{wsa_output};
 
     my @ops  = $wsdl->operations;
     unless(@ops)
@@ -230,13 +299,19 @@ sub operationsFromWSDL($@)
         else
         {   trace __x"add stub handler for operation `{name}'", name => $name;
             my $server  = $op->serverClass;
-            my $handler = $default
+            my $handler = $default_cb
               || sub { $server->faultNotImplemented($name) };
 
             $code = $op->compileHandler(callback => $handler);
         }
 
         $self->addHandler($name, $op, $code);
+
+        if($op->can('wsaAction'))
+        {   $wsa_input->{$name}  ||= $op->wsaAction('INPUT');
+            $wsa_output->{$name} ||= $op->wsaAction('OUTPUT');
+        }
+        $self->addSoapAction($name, $op->soapAction);
     }
 
     info __x"added {nr} operations from WSDL", nr => (scalar @ops);
@@ -283,29 +358,23 @@ sub printIndex(;$)
 
 sub faultInvalidXML($)
 {   my ($self, $error) = @_;
-    my $text = __x"The XML cannot be parsed: {error}", error => $error;
-    (RC_UNPROCESSABLE_ENTITY, 'XML syntax error', $text);
+    ( RC_UNPROCESSABLE_ENTITY, 'XML syntax error'
+    , __x("The XML cannot be parsed: {error}", error => $error));
 }
 
 
 sub faultNotSoapMessage($)
 {   my ($self, $type) = @_;
-
-    my $text =
-        __x"The message was XML, but not SOAP; not an Envelope but `{type}'"
-      , type => $type;
-
-    (RC_FORBIDDEN, 'message not SOAP', $text);
+    ( RC_FORBIDDEN, 'message not SOAP'
+    , __x( "The message was XML, but not SOAP; not an Envelope but `{type}'"
+         , type => $type));
 }
 
 
 sub faultUnsupportedSoapVersion($)
 {   my ($self, $envns) = @_;
-
-    my $text = __x"The soap version `{envns}' is not supported"
-                  , envns => $envns;
-
-    (RC_NOT_IMPLEMENTED, 'SOAP version not supported', $text);
+    ( RC_NOT_IMPLEMENTED, 'SOAP version not supported'
+    , __x("The soap version `{envns}' is not supported", envns => $envns));
 }
 
 
